@@ -38,11 +38,11 @@ pub fn gen_function(def: &MethodDef, gen: &Gen) -> TokenStream {
     }
 }
 
-fn gen_function_if(entry: &[ElementType], gen: &Gen) -> TokenStream {
+fn gen_function_if(entry: &[Type], gen: &Gen) -> TokenStream {
     let mut tokens = TokenStream::new();
 
     for def in entry {
-        if let ElementType::MethodDef(def) = def {
+        if let Type::MethodDef(def) = def {
             tokens.combine(&gen_sys_function(def, gen));
         }
     }
@@ -53,7 +53,9 @@ fn gen_function_if(entry: &[ElementType], gen: &Gen) -> TokenStream {
 fn gen_sys_function(def: &MethodDef, gen: &Gen) -> TokenStream {
     let name = gen_ident(def.name());
     let signature = def.signature(&[]);
-    let cfg = gen.function_cfg(def).gen_with_doc(gen);
+    let cfg = def.cfg();
+    let doc = gen.doc(&cfg);
+    let features = gen.cfg(&cfg);
     let mut return_type = gen_return_sig(&signature, gen);
 
     if return_type.is_empty() {
@@ -61,13 +63,14 @@ fn gen_sys_function(def: &MethodDef, gen: &Gen) -> TokenStream {
     }
 
     let params = signature.params.iter().map(|p| {
-        let name = gen_param_name(&p.param);
-        let tokens = gen_param_sig(p, gen);
+        let name = gen_param_name(&p.def);
+        let tokens = gen_default_type(&p.ty, gen);
         quote! { #name: #tokens }
     });
 
     quote! {
-        #cfg
+        #doc
+        #features
         pub fn #name(#(#params),*) #return_type;
     }
 }
@@ -78,8 +81,8 @@ fn gen_win_function(def: &MethodDef, gen: &Gen) -> TokenStream {
     let constraints = gen_param_constraints(&signature.params, gen);
 
     let abi_params = signature.params.iter().map(|p| {
-        let name = gen_param_name(&p.param);
-        let tokens = gen_abi_param_sig(p, gen);
+        let name = gen_param_name(&p.def);
+        let tokens = gen_abi_element_name(&p.ty, gen);
         quote! { #name: #tokens }
     });
 
@@ -98,16 +101,19 @@ fn gen_win_function(def: &MethodDef, gen: &Gen) -> TokenStream {
         }
     };
 
-    let cfg = gen.function_cfg(def).gen_with_doc(gen);
+    let cfg = def.cfg();
+    let doc = gen.doc(&cfg);
+    let features = gen.cfg(&cfg);
 
     match signature.kind() {
         SignatureKind::Query => {
             let leading_params = &signature.params[..signature.params.len() - 2];
-            let args = leading_params.iter().map(gen_win32_abi_arg);
+            let args = gen_win32_args(leading_params);
             let params = gen_win32_params(leading_params, gen);
 
             quote! {
-                #cfg
+                #doc
+                #features
                 #[inline]
                 pub unsafe fn #name<#constraints T: ::windows::core::Interface>(#params) -> ::windows::core::Result<T> {
                     #[cfg(windows)]
@@ -117,7 +123,7 @@ fn gen_win_function(def: &MethodDef, gen: &Gen) -> TokenStream {
                             fn #name(#(#abi_params),*) #abi_return_type;
                         }
                         let mut result__ = ::core::option::Option::None;
-                        #name(#(#args,)* &<T as ::windows::core::Interface>::IID, &mut result__ as *mut _ as *mut _).and_some(result__)
+                        #name(#args &<T as ::windows::core::Interface>::IID, &mut result__ as *mut _ as *mut _).and_some(result__)
                     }
                     #[cfg(not(windows))]
                     unimplemented!("Unsupported target OS");
@@ -126,11 +132,12 @@ fn gen_win_function(def: &MethodDef, gen: &Gen) -> TokenStream {
         }
         SignatureKind::QueryOptional => {
             let leading_params = &signature.params[..signature.params.len() - 2];
-            let args = leading_params.iter().map(gen_win32_abi_arg);
+            let args = gen_win32_args(leading_params);
             let params = gen_win32_params(leading_params, gen);
 
             quote! {
-                #cfg
+                #doc
+                #features
                 #[inline]
                 pub unsafe fn #name<#constraints T: ::windows::core::Interface>(#params result__: *mut ::core::option::Option<T>) -> ::windows::core::Result<()> {
                     #[cfg(windows)]
@@ -139,7 +146,7 @@ fn gen_win_function(def: &MethodDef, gen: &Gen) -> TokenStream {
                         extern "system" {
                             fn #name(#(#abi_params),*) #abi_return_type;
                         }
-                        #name(#(#args,)* &<T as ::windows::core::Interface>::IID, result__ as *mut _ as *mut _).ok()
+                        #name(#args &<T as ::windows::core::Interface>::IID, result__ as *mut _ as *mut _).ok()
                     }
                     #[cfg(not(windows))]
                     unimplemented!("Unsupported target OS");
@@ -148,15 +155,15 @@ fn gen_win_function(def: &MethodDef, gen: &Gen) -> TokenStream {
         }
         SignatureKind::ResultValue => {
             let leading_params = &signature.params[..signature.params.len() - 1];
-            let args = leading_params.iter().map(gen_win32_abi_arg);
+            let args = gen_win32_args(leading_params);
             let params = gen_win32_params(leading_params, gen);
-            let mut return_sig = signature.params[signature.params.len() - 1].signature.clone();
-            return_sig.pointers -= 1;
-            let return_type_tokens = gen_result_sig(&return_sig, gen);
-            let abi_return_type_tokens = gen_abi_sig(&return_sig, gen);
+            let return_type = signature.params[signature.params.len() - 1].ty.deref();
+            let return_type_tokens = gen_element_name(&return_type, gen);
+            let abi_return_type_tokens = gen_abi_element_name(&return_type, gen);
 
             quote! {
-                #cfg
+                #doc
+                #features
                 #[inline]
                 pub unsafe fn #name<#constraints>(#params) -> ::windows::core::Result<#return_type_tokens> {
                     #[cfg(windows)]
@@ -166,7 +173,7 @@ fn gen_win_function(def: &MethodDef, gen: &Gen) -> TokenStream {
                             fn #name(#(#abi_params),*) #abi_return_type;
                         }
                         let mut result__: #abi_return_type_tokens = ::core::mem::zeroed();
-                        #name(#(#args,)* ::core::mem::transmute(&mut result__)).from_abi::<#return_type_tokens>(result__)
+                        #name(#args ::core::mem::transmute(&mut result__)).from_abi::<#return_type_tokens>(result__)
                     }
                     #[cfg(not(windows))]
                     unimplemented!("Unsupported target OS");
@@ -174,11 +181,12 @@ fn gen_win_function(def: &MethodDef, gen: &Gen) -> TokenStream {
             }
         }
         SignatureKind::ResultVoid => {
+            let args = gen_win32_args(&signature.params);
             let params = gen_win32_params(&signature.params, gen);
-            let args = signature.params.iter().map(gen_win32_abi_arg);
 
             quote! {
-                #cfg
+                #doc
+                #features
                 #[inline]
                 pub unsafe fn #name<#constraints>(#params) -> ::windows::core::Result<()> {
                     #[cfg(windows)]
@@ -187,7 +195,7 @@ fn gen_win_function(def: &MethodDef, gen: &Gen) -> TokenStream {
                         extern "system" {
                             fn #name(#(#abi_params),*) #abi_return_type;
                         }
-                        #name(#(#args),*).ok()
+                        #name(#args).ok()
                     }
                     #[cfg(not(windows))]
                     unimplemented!("Unsupported target OS");
@@ -195,33 +203,60 @@ fn gen_win_function(def: &MethodDef, gen: &Gen) -> TokenStream {
             }
         }
         SignatureKind::ReturnStruct | SignatureKind::PreserveSig => {
-            let params = gen_win32_params(&signature.params, gen);
-            let args = signature.params.iter().map(gen_win32_abi_arg);
+            if handle_last_error(def, &signature) {
+                let args = gen_win32_args(&signature.params);
+                let params = gen_win32_params(&signature.params, gen);
+                let return_type = gen_element_name(&signature.return_type.unwrap(), gen);
 
-            quote! {
-                #cfg
-                #[inline]
-                pub unsafe fn #name<#constraints>(#params) #abi_return_type {
-                    #[cfg(windows)]
-                    {
-                        #link_attr
-                        extern "system" {
-                            fn #name(#(#abi_params),*) #abi_return_type;
+                quote! {
+                    #doc
+                    #features
+                    #[inline]
+                    pub unsafe fn #name<#constraints>(#params) -> ::windows::core::Result<#return_type> {
+                        #[cfg(windows)]
+                        {
+                            #link_attr
+                            extern "system" {
+                                fn #name(#(#abi_params),*) -> #return_type;
+                            }
+                            let result__ = #name(#args);
+                            (!result__.is_invalid()).then(||result__).ok_or_else(::windows::core::Error::from_win32)
                         }
-                        ::core::mem::transmute(#name(#(#args),*))
+                        #[cfg(not(windows))]
+                        unimplemented!("Unsupported target OS");
+                    }
+                }
+            } else {
+                let args = gen_win32_args(&signature.params);
+                let params = gen_win32_params(&signature.params, gen);
+
+                quote! {
+                    #doc
+                    #features
+                    #[inline]
+                    pub unsafe fn #name<#constraints>(#params) #abi_return_type {
+                        #[cfg(windows)]
+                        {
+                            #link_attr
+                            extern "system" {
+                                fn #name(#(#abi_params),*) #abi_return_type;
+                            }
+                            ::core::mem::transmute(#name(#args))
+                        }
+                        #[cfg(not(windows))]
+                        unimplemented!("Unsupported target OS");
                     }
-                    #[cfg(not(windows))]
-                    unimplemented!("Unsupported target OS");
                 }
             }
         }
         SignatureKind::ReturnVoid => {
+            let args = gen_win32_args(&signature.params);
             let params = gen_win32_params(&signature.params, gen);
-            let args = signature.params.iter().map(gen_win32_abi_arg);
             let does_not_return = does_not_return(def);
 
             quote! {
-                #cfg
+                #doc
+                #features
                 #[inline]
                 pub unsafe fn #name<#constraints>(#params) #does_not_return {
                     #[cfg(windows)]
@@ -230,7 +265,7 @@ fn gen_win_function(def: &MethodDef, gen: &Gen) -> TokenStream {
                         extern "system" {
                             fn #name(#(#abi_params),*) #does_not_return;
                         }
-                        #name(#(#args),*)
+                        #name(#args)
                     }
                     #[cfg(not(windows))]
                     unimplemented!("Unsupported target OS");
@@ -247,3 +282,21 @@ fn does_not_return(def: &MethodDef) -> TokenStream {
         quote! {}
     }
 }
+
+fn handle_last_error(def: &MethodDef, signature: &Signature) -> bool {
+    if let Some(map) = def.impl_map() {
+        if map.flags().last_error() {
+            if let Some(Type::TypeDef(return_type)) = &signature.return_type {
+                if return_type.is_handle() {
+                    if return_type.underlying_type().is_pointer() {
+                        return true;
+                    }
+                    if !return_type.invalid_values().is_empty() {
+                        return true;
+                    }
+                }
+            }
+        }
+    }
+    false
+}
