@@ -10,7 +10,7 @@ pub fn gen(def: &TypeDef, gen: &Gen) -> TokenStream {
 
 pub fn gen_sys_handle(def: &TypeDef, gen: &Gen) -> TokenStream {
     let ident = gen_ident(def.name());
-    let signature = gen_signature(def, gen);
+    let signature = gen_default_type(&def.underlying_type(), gen);
 
     quote! {
         pub type #ident = #signature;
@@ -20,26 +20,47 @@ pub fn gen_sys_handle(def: &TypeDef, gen: &Gen) -> TokenStream {
 pub fn gen_win_handle(def: &TypeDef, gen: &Gen) -> TokenStream {
     let name = def.name();
     let ident = gen_ident(def.name());
-    let signature = gen_signature(def, gen);
+    let underlying_type = def.underlying_type();
+    let signature = gen_default_type(&underlying_type, gen);
+    let check = if underlying_type.is_pointer() {
+        quote! {
+            impl #ident {
+                pub fn is_invalid(&self) -> bool {
+                    self.0.is_null()
+                }
+            }
+        }
+    } else {
+        let invalid = def.invalid_values();
+
+        if !invalid.is_empty() {
+            let invalid = invalid.iter().map(|value| {
+                let literal = Literal::i64_unsuffixed(*value);
+
+                if *value < 0 && underlying_type.is_unsigned() {
+                    quote! { self.0 == #literal as _ }
+                } else {
+                    quote! { self.0 == #literal }
+                }
+            });
+            quote! {
+                impl #ident {
+                    pub fn is_invalid(&self) -> bool {
+                        #(#invalid)||*
+                    }
+                }
+            }
+        } else {
+            quote! {}
+        }
+    };
 
     let mut tokens = quote! {
         #[repr(transparent)]
         // Unfortunately, Rust requires these to be derived to allow constant patterns.
         #[derive(::core::cmp::PartialEq, ::core::cmp::Eq)]
         pub struct #ident(pub #signature);
-        impl #ident {
-            pub fn is_invalid(&self) -> bool {
-                *self == unsafe { ::core::mem::zeroed() }
-            }
-
-            pub fn ok(self) -> ::windows::core::Result<Self> {
-                if !self.is_invalid() {
-                    Ok(self)
-                } else {
-                    Err(::windows::core::Error::from_win32())
-                }
-            }
-        }
+        #check
         impl ::core::default::Default for #ident {
             fn default() -> Self {
                 unsafe { ::core::mem::zeroed() }
@@ -77,12 +98,3 @@ pub fn gen_win_handle(def: &TypeDef, gen: &Gen) -> TokenStream {
 
     tokens
 }
-
-fn gen_signature(def: &TypeDef, gen: &Gen) -> TokenStream {
-    let mut signature = def.fields().next().map(|field| field.signature(Some(def))).unwrap();
-    // Handle fields don't need to be mutable and it's easier to work with handles when
-    // they're `*const` rather than `*mut`. This mainly affects string handle types
-    // like PWSTR/PSTR/BSTR.
-    signature.is_const = true;
-    gen_sig(&signature, gen)
-}
