@@ -5,7 +5,7 @@ pub fn gen_std_traits(def: &TypeDef, cfg: &Cfg, gen: &Gen) -> TokenStream {
     let name = ident.as_str();
     let constraints = gen_type_constraints(def, gen);
     let phantoms = gen_phantoms(def, gen);
-    let cfg = cfg.gen(gen);
+    let cfg = gen.cfg(cfg);
 
     quote! {
         #cfg
@@ -32,7 +32,7 @@ pub fn gen_std_traits(def: &TypeDef, cfg: &Cfg, gen: &Gen) -> TokenStream {
 }
 
 pub fn gen_interface_trait(def: &TypeDef, cfg: &Cfg, gen: &Gen) -> TokenStream {
-    let cfg = cfg.gen(gen);
+    let cfg = gen.cfg(cfg);
     if let Some(default) = def.default_interface() {
         let name = gen_type_ident(def, gen);
         let default_name = gen_type_ident(&default, gen);
@@ -63,7 +63,7 @@ pub fn gen_interface_trait(def: &TypeDef, cfg: &Cfg, gen: &Gen) -> TokenStream {
 }
 
 pub fn gen_runtime_trait(def: &TypeDef, cfg: &Cfg, gen: &Gen) -> TokenStream {
-    let cfg = cfg.gen(gen);
+    let cfg = gen.cfg(cfg);
     if def.is_winrt() {
         let name = gen_type_ident(def, gen);
         let constraints = gen_type_constraints(def, gen);
@@ -78,6 +78,10 @@ pub fn gen_runtime_trait(def: &TypeDef, cfg: &Cfg, gen: &Gen) -> TokenStream {
             #cfg
             unsafe impl<#(#constraints)*> ::windows::core::RuntimeType for #name {
                 const SIGNATURE: ::windows::core::ConstBuffer = #type_signature;
+                type DefaultType = ::core::option::Option<Self>;
+                fn from_default(from: &Self::DefaultType) -> ::windows::core::Result<Self> {
+                    from.as_ref().cloned().ok_or(::windows::core::Error::OK)
+                }
             }
         }
     } else {
@@ -88,25 +92,26 @@ pub fn gen_runtime_trait(def: &TypeDef, cfg: &Cfg, gen: &Gen) -> TokenStream {
 pub fn gen_vtbl_signature(def: &TypeDef, method: &MethodDef, gen: &Gen) -> TokenStream {
     let is_winrt = def.is_winrt();
     let signature = method.signature(&def.generics);
-    let hresult = gen_element_name(&ElementType::HRESULT, gen);
+    let hresult = gen_element_name(&Type::HRESULT, gen);
 
     let (trailing_return_type, return_type, udt_return_type) = if is_winrt {
-        if let Some(return_sig) = &signature.return_sig {
-            let tokens = gen_abi_sig(return_sig, gen);
-            if return_sig.is_array {
+        if let Some(return_type) = &signature.return_type {
+            if let Type::WinrtArray(kind) = return_type {
+                let tokens = gen_abi_element_name(kind, gen);
                 (quote! { result_size__: *mut u32, result__: *mut *mut #tokens }, quote! { -> #hresult }, quote! {})
             } else {
+                let tokens = gen_abi_element_name(return_type, gen);
                 (quote! { result__: *mut #tokens }, quote! { -> #hresult }, quote! {})
             }
         } else {
             (quote! {}, quote! { -> #hresult }, quote! {})
         }
-    } else if let Some(return_sig) = &signature.return_sig {
-        if return_sig.is_udt() {
-            let tokens = gen_abi_sig(return_sig, gen);
+    } else if let Some(return_type) = &signature.return_type {
+        if return_type.is_udt() {
+            let tokens = gen_abi_element_name(return_type, gen);
             (quote! {}, quote! {}, quote! { result__: *mut #tokens, })
         } else {
-            let tokens = gen_sig(return_sig, gen);
+            let tokens = gen_default_type(return_type, gen);
             (quote! {}, quote! { -> #tokens }, quote! {})
         }
     } else {
@@ -114,32 +119,28 @@ pub fn gen_vtbl_signature(def: &TypeDef, method: &MethodDef, gen: &Gen) -> Token
     };
 
     let params = signature.params.iter().map(|p| {
-        let name = gen_param_name(&p.param);
+        let name = gen_param_name(&p.def);
         if is_winrt {
-            let signature = &p.signature;
-            let abi = gen_abi_sig(signature, gen);
+            let abi = gen_abi_element_name(&p.ty, gen);
+            let abi_size_name = gen_ident(&format!("{}_array_size", p.def.name()));
 
-            if signature.is_array {
-                let abi_size_name = gen_ident(&format!("{}_array_size", p.param.name()));
-
-                if p.param.is_input() {
+            if p.def.flags().input() {
+                if p.ty.is_winrt_array() {
                     quote! { #abi_size_name: u32, #name: *const #abi, }
-                } else if p.signature.by_ref {
-                    quote! { #abi_size_name: *mut u32, #name: *mut *mut #abi, }
-                } else {
-                    quote! { #abi_size_name: u32, #name: *mut #abi, }
-                }
-            } else if p.param.is_input() {
-                if p.signature.is_const {
+                } else if p.ty.is_winrt_const_ref() {
                     quote! { #name: &#abi, }
                 } else {
                     quote! { #name: #abi, }
                 }
+            } else if p.ty.is_winrt_array() {
+                quote! { #abi_size_name: u32, #name: *mut #abi, }
+            } else if p.ty.is_winrt_array_ref() {
+                quote! { #abi_size_name: *mut u32, #name: *mut *mut #abi, }
             } else {
                 quote! { #name: *mut #abi, }
             }
         } else {
-            let abi = gen_abi_param_sig(p, gen);
+            let abi = gen_abi_element_name(&p.ty, gen);
             quote! { #name: #abi, }
         }
     });
@@ -148,7 +149,7 @@ pub fn gen_vtbl_signature(def: &TypeDef, method: &MethodDef, gen: &Gen) -> Token
 }
 
 pub fn gen_vtbl(def: &TypeDef, cfg: &Cfg, gen: &Gen) -> TokenStream {
-    let cfg = cfg.gen(gen);
+    let cfg = gen.cfg(cfg);
     let vtbl = gen_vtbl_ident(def, gen);
     let phantoms = gen_named_phantoms(def, gen);
     let constraints = gen_type_constraints(def, gen);
@@ -157,12 +158,12 @@ pub fn gen_vtbl(def: &TypeDef, cfg: &Cfg, gen: &Gen) -> TokenStream {
     method_names.add_vtable_types(def);
 
     match def.vtable_types().last() {
-        Some(ElementType::IUnknown) => methods.combine(&quote! { pub base: ::windows::core::IUnknownVtbl, }),
-        Some(ElementType::IInspectable) => methods.combine(&quote! { pub base: ::windows::core::IInspectableVtbl, }),
-        Some(ElementType::TypeDef(def)) => {
+        Some(Type::IUnknown) => methods.combine(&quote! { pub base__: ::windows::core::IUnknownVtbl, }),
+        Some(Type::IInspectable) => methods.combine(&quote! { pub base__: ::windows::core::IInspectableVtbl, }),
+        Some(Type::TypeDef(def)) => {
             let vtbl = gen_vtbl_ident(def, gen);
             let namespace = gen.namespace(def.namespace());
-            methods.combine(&quote! { pub base: #namespace #vtbl, });
+            methods.combine(&quote! { pub base__: #namespace #vtbl, });
         }
         _ => {}
     }
@@ -174,9 +175,10 @@ pub fn gen_vtbl(def: &TypeDef, cfg: &Cfg, gen: &Gen) -> TokenStream {
 
         let name = method_names.add(&method);
         let signature = gen_vtbl_signature(def, &method, gen);
-        let cfg = gen.method_cfg(def, &method);
-        let cfg_all = cfg.gen(gen);
-        let cfg_not = cfg.gen_not(gen);
+        let mut cfg = method.cfg();
+        cfg.add_feature(def.namespace());
+        let cfg_all = gen.cfg(&cfg);
+        let cfg_not = gen.not_cfg(&cfg);
 
         let signature = quote! { pub #name: unsafe extern "system" fn #signature, };
 
@@ -251,7 +253,7 @@ pub fn gen_constant_type_value(value: &ConstantValue) -> TokenStream {
 }
 
 pub fn gen_guid(value: &GUID, gen: &Gen) -> TokenStream {
-    let guid = gen_element_name(&ElementType::GUID, gen);
+    let guid = gen_element_name(&Type::GUID, gen);
 
     if gen.sys {
         let a = Literal::u32_unsuffixed(value.0);
@@ -296,7 +298,7 @@ pub fn gen_constant_value(value: &ConstantValue) -> TokenStream {
 
 pub fn gen_runtime_name(def: &TypeDef, cfg: &Cfg, gen: &Gen) -> TokenStream {
     let name = gen_type_ident(def, gen);
-    let cfg = cfg.gen(gen);
+    let cfg = gen.cfg(cfg);
 
     if def.is_winrt() {
         let constraints = gen_type_constraints(def, gen);
@@ -308,7 +310,7 @@ pub fn gen_runtime_name(def: &TypeDef, cfg: &Cfg, gen: &Gen) -> TokenStream {
                 const NAME: &'static str = #runtime_name;
             }
         }
-    } else if def.vtable_types().iter().any(|e| e == &ElementType::IInspectable) {
+    } else if def.vtable_types().iter().any(|e| e == &Type::IInspectable) {
         quote! {
             #cfg
             impl ::windows::core::RuntimeName for #name {
@@ -320,12 +322,12 @@ pub fn gen_runtime_name(def: &TypeDef, cfg: &Cfg, gen: &Gen) -> TokenStream {
     }
 }
 
-pub fn gen_win32_upcall(sig: &MethodSignature, inner: TokenStream) -> TokenStream {
+pub fn gen_win32_upcall(sig: &Signature, inner: TokenStream) -> TokenStream {
     match sig.kind() {
         SignatureKind::ResultValue => {
             let invoke_args = sig.params[..sig.params.len() - 1].iter().map(gen_win32_invoke_arg);
 
-            let result = gen_param_name(&sig.params[sig.params.len() - 1].param);
+            let result = gen_param_name(&sig.params[sig.params.len() - 1].def);
 
             quote! {
                 match #inner(#(#invoke_args,)*) {
@@ -361,11 +363,11 @@ pub fn gen_win32_upcall(sig: &MethodSignature, inner: TokenStream) -> TokenStrea
     }
 }
 
-pub fn gen_winrt_upcall(sig: &MethodSignature, inner: TokenStream) -> TokenStream {
+pub fn gen_winrt_upcall(sig: &Signature, inner: TokenStream) -> TokenStream {
     let invoke_args = sig.params.iter().map(gen_winrt_invoke_arg);
 
-    match &sig.return_sig {
-        Some(return_sig) if return_sig.is_array => {
+    match &sig.return_type {
+        Some(return_type) if return_type.is_winrt_array() => {
             quote! {
                 match #inner(#(#invoke_args,)*) {
                     ::core::result::Result::Ok(ok__) => {
@@ -397,9 +399,9 @@ pub fn gen_winrt_upcall(sig: &MethodSignature, inner: TokenStream) -> TokenStrea
 }
 
 fn gen_win32_invoke_arg(param: &MethodParam) -> TokenStream {
-    let name = gen_param_name(&param.param);
+    let name = gen_param_name(&param.def);
 
-    if param.signature.pointers == 0 && param.signature.kind.is_nullable() {
+    if (!param.ty.is_pointer() && param.ty.is_nullable()) || (param.def.flags().input() && !param.ty.is_primitive()) {
         quote! { ::core::mem::transmute(&#name) }
     } else {
         quote! { ::core::mem::transmute_copy(&#name) }
@@ -407,26 +409,23 @@ fn gen_win32_invoke_arg(param: &MethodParam) -> TokenStream {
 }
 
 fn gen_winrt_invoke_arg(param: &MethodParam) -> TokenStream {
-    let name = gen_param_name(&param.param);
+    let name = gen_param_name(&param.def);
+    let abi_size_name: TokenStream = format!("{}_array_size", param.def.name()).into();
 
-    if param.signature.is_array {
-        let abi_size_name: TokenStream = format!("{}_array_size", param.param.name()).into();
-
-        if param.param.is_input() {
+    if param.def.flags().input() {
+        if param.ty.is_winrt_array() {
             quote! { ::core::slice::from_raw_parts(::core::mem::transmute_copy(&#name), #abi_size_name as _) }
-        } else if param.signature.by_ref {
-            quote! { ::windows::core::ArrayProxy::from_raw_parts(::core::mem::transmute_copy(&#name), #abi_size_name).as_array() }
-        } else {
-            quote! { ::core::slice::from_raw_parts_mut(::core::mem::transmute_copy(&#name), #abi_size_name as _) }
-        }
-    } else if param.param.is_input() {
-        if param.signature.kind.is_primitive() {
+        } else if param.ty.is_primitive() {
             quote! { #name }
-        } else if param.signature.is_const {
+        } else if param.ty.is_winrt_const_ref() {
             quote! { ::core::mem::transmute_copy(&#name) }
         } else {
             quote! { ::core::mem::transmute(&#name) }
         }
+    } else if param.ty.is_winrt_array() {
+        quote! { ::core::slice::from_raw_parts_mut(::core::mem::transmute_copy(&#name), #abi_size_name as _) }
+    } else if param.ty.is_winrt_array_ref() {
+        quote! { ::windows::core::ArrayProxy::from_raw_parts(::core::mem::transmute_copy(&#name), #abi_size_name).as_array() }
     } else {
         quote! { ::core::mem::transmute_copy(&#name) }
     }
@@ -439,10 +438,10 @@ pub fn gen_impl_signature(def: &TypeDef, method: &MethodDef, gen: &Gen) -> Token
         let is_delegate = def.kind() == TypeKind::Delegate;
         let params = signature.params.iter().map(|p| gen_winrt_produce_type(p, !is_delegate, gen));
 
-        let return_sig = if let Some(return_sig) = &signature.return_sig {
-            let tokens = gen_element_name(&return_sig.kind, gen);
+        let return_type = if let Some(return_type) = &signature.return_type {
+            let tokens = gen_element_name(return_type, gen);
 
-            if return_sig.is_array {
+            if return_type.is_winrt_array() {
                 quote! { ::windows::core::Array<#tokens> }
             } else {
                 tokens
@@ -454,10 +453,10 @@ pub fn gen_impl_signature(def: &TypeDef, method: &MethodDef, gen: &Gen) -> Token
         let this = if is_delegate {
             quote! {}
         } else {
-            quote! { &mut self, }
+            quote! { &self, }
         };
 
-        quote! { (#this #(#params),*) -> ::windows::core::Result<#return_sig> }
+        quote! { (#this #(#params),*) -> ::windows::core::Result<#return_type> }
     } else {
         let signature_kind = signature.kind();
         let mut params = quote! {};
@@ -472,73 +471,71 @@ pub fn gen_impl_signature(def: &TypeDef, method: &MethodDef, gen: &Gen) -> Token
             }
         }
 
-        let return_sig = match signature_kind {
+        let return_type = match signature_kind {
             SignatureKind::ReturnVoid => quote! {},
             SignatureKind::Query | SignatureKind::QueryOptional | SignatureKind::ResultVoid => quote! { -> ::windows::core::Result<()> },
             SignatureKind::ResultValue => {
-                let mut return_sig = signature.params[signature.params.len() - 1].signature.clone();
-                return_sig.pointers -= 1;
-                let return_sig = gen_result_sig(&return_sig, gen);
+                let return_type = signature.params[signature.params.len() - 1].ty.deref();
+                let return_type = gen_element_name(&return_type, gen);
 
-                quote! { -> ::windows::core::Result<#return_sig> }
+                quote! { -> ::windows::core::Result<#return_type> }
             }
             _ => gen_return_sig(&signature, gen),
         };
 
-        quote! { (&mut self, #params) #return_sig }
+        quote! { (&self, #params) #return_type }
     }
 }
 
 fn gen_win32_produce_type(param: &MethodParam, gen: &Gen) -> TokenStream {
-    let name = gen_param_name(&param.param);
-    let kind = gen_param_sig(param, gen);
+    let name = gen_param_name(&param.def);
+    let kind = gen_default_type(&param.ty, gen);
 
-    if param.param.is_input() && !param.signature.is_primitive() {
+    if param.def.flags().input() && !param.ty.is_primitive() {
         quote! { #name: &#kind, }
     } else {
         quote! { #name: #kind, }
     }
 }
 
-fn gen_winrt_produce_type(param: &MethodParam, include_param_names: bool, gen: &Gen) -> TokenStream {
-    let kind = gen_element_name(&param.signature.kind, gen);
-
-    let sig = if param.signature.is_array {
-        if param.signature.by_ref {
-            quote! { &mut ::windows::core::Array<#kind> }
-        } else {
-            let kind = if let ElementType::GenericParam(_) = param.signature.kind {
-                quote! { <#kind as ::windows::core::DefaultType>::DefaultType }
-            } else if param.signature.kind.is_nullable() {
-                quote! { ::core::option::Option<#kind> }
-            } else {
-                kind
-            };
-
-            if param.param.is_input() {
-                quote! { &[#kind] }
-            } else {
-                quote! { &mut [#kind] }
-            }
-        }
-    } else if param.param.is_input() {
-        if let ElementType::GenericParam(_) = param.signature.kind {
-            quote! { &<#kind as ::windows::core::DefaultType>::DefaultType }
-        } else if param.signature.kind.is_primitive() {
-            quote! { #kind }
-        } else if param.signature.kind.is_nullable() {
-            quote! { &::core::option::Option<#kind> }
-        } else {
-            quote! { &#kind }
-        }
-    } else if param.signature.kind.is_nullable() {
-        quote! { &mut ::core::option::Option<#kind> }
+pub fn gen_default_type(def: &Type, gen: &Gen) -> TokenStream {
+    if let Type::WinrtArray(def) = def {
+        gen_default_type(def, gen)
     } else {
-        quote! { &mut #kind }
+        let kind = gen_element_name(def, gen);
+
+        if def.is_generic() {
+            quote! { <#kind as ::windows::core::RuntimeType>::DefaultType }
+        } else if def.is_nullable() && !gen.sys {
+            quote! { ::core::option::Option<#kind> }
+        } else {
+            kind
+        }
+    }
+}
+
+fn gen_winrt_produce_type(param: &MethodParam, include_param_names: bool, gen: &Gen) -> TokenStream {
+    let default_type = gen_default_type(&param.ty, gen);
+
+    let sig = if param.def.flags().input() {
+        if param.ty.is_winrt_array() {
+            quote! { &[#default_type] }
+        } else if param.ty.is_primitive() {
+            quote! { #default_type }
+        } else {
+            quote! { &#default_type }
+        }
+    } else if param.ty.is_winrt_array() {
+        quote! { &mut [#default_type] }
+    } else if param.ty.is_winrt_array_ref() {
+        let kind = gen_element_name(&param.ty, gen);
+        quote! { &mut ::windows::core::Array<#kind> }
+    } else {
+        quote! { &mut #default_type }
     };
 
     if include_param_names {
-        let name = gen_param_name(&param.param);
+        let name = gen_param_name(&param.def);
         quote! { #name: #sig }
     } else {
         sig
