@@ -26,12 +26,7 @@ pub fn gen() -> TokenStream {
                     return Self(::core::ptr::null_mut());
                 }
 
-                unsafe {
-                    SysAllocStringLen(
-                        PWSTR(value.as_ptr()),
-                        value.len() as u32,
-                    )
-                }
+                unsafe { SysAllocStringLen(value) }
             }
 
             pub fn as_wide(&self) -> &[u16] {
@@ -41,6 +36,14 @@ pub fn gen() -> TokenStream {
 
                 unsafe { ::core::slice::from_raw_parts(self.0, self.len()) }
             }
+
+            pub unsafe fn from_raw(raw: *const u16) -> Self {
+                Self(raw)
+            }
+
+            pub fn into_raw(self) -> *const u16 {
+                unsafe { std::mem::transmute(self) }
+            }
         }
         impl ::core::clone::Clone for BSTR {
             fn clone(&self) -> Self {
